# -*- coding: utf-8 -*-
"""Intel hex functions.

This module provides functions to handle ihex file.
Include parse, format check and others.

Ihex file means the file which is in "Intel Hex Format".
The ihex is common use to store program data. And many mcu
loader is supprot ihex file. The output binary data of compiler
(e.g. GCC, ARMCC, IAR) is also ihex in fefault.

.. _Intel Hexadecimal Object File Format Specification:
    http://www.interlog.com/~speff/usefulinfo/Hexfrmt.pdf
"""

from serprog import exceptions

def parse(filename: str) -> list:
    """ Parse ihex file to data blocks.
    
    Args:
        filename (str): The file to parse.
    
    Raises:
        serprog.exceptions.IhexFormatError:
    
    Returns:
        list: data blocks with start address.
    """

    # IHEX record types
    DATA_RECORD = 0
    EOF_RECORD = 1
    EXT_SEG_ADDR_RECORD = 2
    START_SEG_ADDR_RECORD = 3
    EXT_LINEAR_ADDR_RECORD = 4
    START_LINEAR_ADDR_RECORD = 5

    with open(filename, 'r') as hexfile:
        sections = []
        section_index = 0
        extended_address = 0
        eof_flag = False

        for line in hexfile.readlines():
            # line is end up with '\n'
            if len(line) < 12:
                # less than minimum length of General Record Format
                raise exceptions.IhexFormatError(filename)

            # Check record mark
            if not line.startswith(':'):
                raise exceptions.IhexFormatError(filename)
            
            # Parse IHEX record fields
            record_length = int(line[1:3], 16)
            address = int(line[3:7], 16)
            record_type = int(line[7:9], 16)
            checksum = int(line[-3:-1], 16)

            # check record length and parse data
            if record_length != 0:
                if len(line) != 12 + record_length * 2:
                    raise exceptions.IhexFormatError(filename)
                else:
                    data = bytearray.fromhex(line[9 : 9 + record_length*2])
            else:
                data = b''

            if record_type == DATA_RECORD:
                # Data record
                if section_index == 0:
                    # First section
                    sections.append({
                        'address': (extended_address << 16) + address,
                        'data': data
                    })
                    section_index += 1

                elif (extended_address << 16) + address == sections[section_index-1]['address'] + len(sections[section_index-1]['data']):
                    # Add data to previous section
                    sections[section_index-1]['data'] += data

                else:
                    # Start new section
                    sections.append({
                        'address': (extended_address << 16) + address,
                        'data': data
                    })
                    section_index += 1

            elif record_type == EOF_RECORD:
                # End of file record
                if address == 0:
                    eof_flag = True
                else:
                    raise exceptions.IhexFormatError(filename)

            elif record_type == EXT_SEG_ADDR_RECORD:
                # Extended Segment Address
                pass

            elif record_type == START_SEG_ADDR_RECORD:
                # Start Segment Address
                pass

            elif record_type == EXT_LINEAR_ADDR_RECORD:
                # Extended Linear Address
                extended_address = int(line[9:13], 16)
                pass

            elif record_type == START_LINEAR_ADDR_RECORD:
                # Start Linear Address
                pass

        if not eof_flag:
            raise exceptions.IhexFormatError(filename)

    return sections


def padding_space(h, pgsz: int, space_data: bytes) -> list: 
    """Padding each data block with `space_data` to let block size fit pgsz * N.
    
    Args:
        h (list): response from `serprog.ihex.parse`.
        pgsz (int): page size, e.g. 256, 512.
        space_data (bytes): the byte data used to padding.
    
    Returns:
        list: data blocks
    """
    res = []
    for sect in h:
        sect_addr, sect_data = sect['address'], sect['data']

        # (start address) not equal (pgsz * N)
        # Pad before the data
        start_padding_len = sect_addr % pgsz
        start_padding = bytes(space_data * start_padding_len)
        sect_data = start_padding + sect_data

        # (end address + 1) not equal (pgsz * N)
        # Pad after the data
        end_padding_len = pgsz - len(sect_data) % pgsz if len(sect_data) % pgsz != 0 else 0
        end_padding = bytes(space_data * end_padding_len)
        sect_data = sect_data + end_padding

        res.append({'address': (sect_addr // pgsz) * pgsz, 'data': sect_data})

    return res


def cut_to_pages(h, pgsz):
    """Cut each data block to pages.
    
    Args:
        h (list): response from `serprog.ihex.padding_space`.
        pgsz (int): page size, e.g. 256, 512.
    
    Returns:
        list: data pages
    """
    res = []
    for sect in h:
        sect_addr, sect_data = sect['address'], sect['data']
        for i in range(0, len(sect_data), pgsz):
            page = {'address': sect_addr + i, 'data': sect_data[i : i + pgsz]}
            res.append(page)
    return res


def is_ihex(filename: str) -> bool:
    """Check the file is ihex format.
    
    Args:
        filename (str): The file to check.
    
    Returns:
        bool: Ture or False.
    """
    try:
        parse(filename)
    except exceptions.IhexFormatError:
        return False
    return True



# -*- coding: utf-8 -*-
"""
Device list and the associated function.
"""

device_list = [
    {
        'name': 'auto',
        'dev_type': 0,
        'protocol_version': 0,
        'userapp_start': 0,
        'userapp_size':  0,
        'note': 'Default, auto detect device type.'
    },
    {
        'name': 'ATSAME54_DEVB',
        'dev_type': 1,
        'protocol_version': 1,
        'userapp_start': 0x00010000,
        'userapp_size':  0x000F0000,
        'note': ''
    },
    {
        'name': 'NUM487KM_DEVB',
        'dev_type': 2,
        'protocol_version': 1,
        'userapp_start': 0x00010000,
        'userapp_size':  0x000F0000,
        'note': ''
    }
]

def get_device_by_str(s: str) -> int:
    """ Get the device number.

    Args:
        s (str): String. Device name, device number are accepted.

    Returns:
        int: Device number. If returned -1, it failed. (wrong format s)
    """
    if s.isdigit():
        for d in device_list:
            if int(s) == d['dev_type']:
                return d['dev_type']
    else:
        for d in device_list:
            if s == d['name']:
                return d['dev_type']
    return -1



# -*- coding: utf-8 -*-
class Error(Exception):
    pass

class IhexFormatError(Error):
    filename: str
    def __init__(self, filename):
        self.filename = filename

class FlashIsNotIhexError(Error):
    filename: str
    def __init__(self, filename):
        self.filename = filename

class EepromIsNotIhexError(Error):
    filename: str
    def __init__(self, filename):
        self.filename = filename

class DeviceTypeError(Error):
    device_type: int
    def __init__(self, device_type):
        self.device_type = device_type

class DeviceNotResponseError(Error):
    def __init__(self):
        pass

class ComuError(Error):
    def __init__(self):
        pass

class CheckDeviceError(Error):
    in_dev: int
    real_dev: int
    def __init__(self, in_dev, real_dev):
        self.in_dev = in_dev
        self.real_dev = real_dev



# -*- coding: utf-8 -*-
"""
Bootloader Protocol implementation.
"""
import enum
from typing import Union

HEADER = b'\xA5\xA5\xA5'

class CMD(enum.IntEnum):
    CHK_PROTOCOL            = 0x01
    CHK_DEVICE              = 0x02
    PROG_END                = 0x03
    PROG_EXT_FLASH_BOOT     = 0x04

    FLASH_SET_PGSZ          = 0x10
    FLASH_GET_PGSZ          = 0x11
    FLASH_WRITE             = 0x12
    FLASH_READ              = 0x13
    FLASH_VERIFY            = 0x14
    FLASH_EARSE_SECTOR      = 0x15
    FLASH_EARSE_ALL         = 0x16

    EEPROM_SET_PGSZ         = 0x20
    EEPROM_GET_PGSZ         = 0x21
    EEPROM_WRITE            = 0x22
    EEPROM_READ             = 0x23
    EEPROM_EARSE            = 0x24
    EEPROM_EARSE_ALL        = 0x25

    EXT_FLASH_FOPEN         = 0x30
    EXT_FLASH_FCLOSE        = 0x31
    EXT_FLASH_WRITE         = 0x32
    EXT_FLASH_READ          = 0x33
    EXT_FLASH_VERIFY        = 0x34
    EXT_FLASH_EARSE_SECTOR  = 0x35
    EXT_FLASH_HEX_DEL       = 0x36

class Decoder(object):
    class _Status(enum.IntEnum):
        HEADER  = 0
        COMMAND = 1
        LENGTH  = 2
        DATA    = 3
        CHKSUM  = 4

    _header_buffer = b'\x00\x00\x00'
    _counter = int()
    _command = int()
    _length  = int()
    _data    = bytes()
    _chksum  = int()
    _isError = bool()
    _isDone  = bool()

    def __init__(self):
        super(Decoder, self).__init__()
        self._status = self._Status(self._Status.HEADER)

    def step(self, ch):

        if self._status is self._Status.HEADER:
            self._header_buffer = self._header_buffer[1:3] + bytes([ch])
            if self._header_buffer == HEADER:
                self._chksum = 0
                self._status = self._Status.COMMAND

        elif self._status is self._Status.COMMAND:
            self._command = CMD(ch)
            self._counter = 0
            self._status = self._Status.LENGTH

        elif self._status is self._Status.LENGTH:
            self._counter = self._counter + 1
            if self._counter == 1:
                self._length = ch << 8
            elif self._counter == 2:
                self._length += ch
                self._counter = 0
                self._data = b''
                if self._length == 0:
                    self._status = self._Status.CHKSUM
                else:
                    self._status = self._Status.DATA

        elif self._status is self._Status.DATA:
            self._chksum += ch
            self._counter = self._counter + 1
            self._data += bytes([ch])
            if self._counter == self._length:
                self._status = self._Status.CHKSUM

        elif self._status is self._Status.CHKSUM:
            if self._chksum % 256 != ch:
                self._error = True
            self._status = self._Status.HEADER
            self._header_buffer = b'\x00\x00\x00'
            self._isDone = True

    def isDone(self):
        return self._isDone

    def isError(self):
        return self._isError

    def getPacket(self):
        if self.isDone():
            res = {
                'command': self._command,
                'data': self._data
            }
            self._isDone = False
        else:
            res = {
                'command': None,
                'data': b''
            }
        return res

def encode(cmd: Union[int, CMD], data: bytes) -> bytes:
    """Encode command, data to a package.
    
    Args:
        cmd (Union[int, Command]): Command in the package.
        data (bytes): Data in the package.
    
    Returns:
        bytes: Package as bytes.
    """

    res  = bytes()
    res += HEADER
    res += cmd.to_bytes(1, 'little')
    res += len(data).to_bytes(2, 'big')
    res += data
    res += (sum(data) % 256).to_bytes(1, 'little')
    return res



# -*- coding: utf-8 -*-

# import conftest

from serprog import ihex
# from distutils import dir_util
# import pytest
# import os


# @pytest.fixture
# def datadir(tmpdir, request):
#     '''
#     Fixture responsible for searching a folder with the same name of test
#     module and, if available, moving all contents to a temporary directory so
#     tests can use them freely.
#     '''
#     filename = request.module.__file__
#     test_dir, _ = os.path.splitext(filename)

#     if os.path.isdir(test_dir):
#         dir_util.copy_tree(test_dir, str(tmpdir))

#     return tmpdir


def test_ihex_parse_1():
    """測試分析單一區段ihex檔案
    """
    predict = [
        {
            'address': 0x0000,
            'data': b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0A\x0B\x0C\x0D\x0E\x0F'
        }
    ]

    # test_ihex_file = datadir.join('ihex1.hex')
    real = ihex.parse('test/ihex1.hex')
    # print(real)
    assert(real == predict)


def test_ihex_parse_2():
    """測試分析雙區段ihex檔案
    """
    predict = [
        {
            'address': 0x0000,
            'data': b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0A\x0B\x0C\x0D\x0E\x0F'
        },
        {
            'address': 0x0100,
            'data': b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0A\x0B\x0C\x0D\x0E\x0F'
        }
    ]

    # test_ihex_file = datadir.join('ihex2.hex')
    real = ihex.parse('test/ihex2.hex')
    # print(real)
    assert(real == predict)


def test_ihex_parse_3():
    """測試分析有擴展位址ihex檔案
    第一段無擴展，高2位元組為0x0000
    第二段有擴展，高2位元組為0xABCD
    """
    predict = [
        {
            'address': 0x00000000,
            'data': b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0A\x0B\x0C\x0D\x0E\x0F'
        },
        {
            'address': 0xABCD0100,
            'data': b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0A\x0B\x0C\x0D\x0E\x0F'
        }
    ]

    # test_ihex_file = datadir.join('ihex3.hex')
    real = ihex.parse('test/ihex3.hex')
    # print(real)
    assert(real == predict)


def test_ihex_padding_1():
    """測試補其空白函式，補其單頁
    """
    input = [
        {
            'address': 0,
            'data': b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0A\x0B\x0C\x0D\x0E\x0F'
        },
        {
            'address': 0xABCD0010,
            'data': b'\x01\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0A\x0B\x0C\x0D\x0E\x0F'
        }
    ]
    space_data = bytes.fromhex('FF')
    page_size = 256

    space_padding = b'\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF'
    predict = [
        {
            'address': 0,
            'data': b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0A\x0B\x0C\x0D\x0E\x0F' + space_padding * 15
        },
        {
            'address': 0xABCD0000,
            'data': space_padding + b'\x01\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0A\x0B\x0C\x0D\x0E\x0F' + space_padding * 14
        }
    ]

    real = ihex.padding_space(input, page_size, space_data)
    assert(real == predict)


def test_ihex_padding_2():
    """測試補其空白函式，補其多頁
    """
    input = [
        {
            'address': 3,
            'data': b'\x03\x04\x05\x06\x07\x08\x09\x0A\x0B\x0C'
        }
    ]
    space_data = bytes.fromhex('FF')
    page_size = 4

    predict = [
        {
            'address': 0,
            'data': b'\xFF\xFF\xFF\x03\x04\x05\x06\x07\x08\x09\x0A\x0B\x0C\xFF\xFF\xFF'
        }
    ]

    real = ihex.padding_space(input, page_size, space_data)
    assert(real == predict)


def test_ihex_cut_1():
    """測試切割頁面函式
        NOTE: 此函式輸入資料須為經過padding之codeblock
              大小須符合 pgsz * N
    """

    input = [
        {
            'address': 0,
            'data': b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0A\x0B\x0C\x0D\x0E\x0F'
        },
        {
            'address': 0xABCD0010,
            'data': b'\x01\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0A\x0B\x0C\x0D\x0E\x0F'
        }
    ]
    space_data = bytes.fromhex('FF')
    page_size = 4

    space_padding = b'\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF'
    predict = [
        {
            'address': 0,
            'data': b'\x00\x01\x02\x03'
        },
        {
            'address': 4,
            'data': b'\x04\x05\x06\x07'
        },
        {
            'address': 8,
            'data': b'\x08\x09\x0A\x0B'
        },
        {
            'address': 12,
            'data': b'\x0C\x0D\x0E\x0F'
        },
        {
            'address': 0xABCD0010,
            'data': b'\x01\x01\x02\x03'
        },
        {
            'address': 0xABCD0014,
            'data': b'\x04\x05\x06\x07'
        },
        {
            'address': 0xABCD0018,
            'data': b'\x08\x09\x0A\x0B'
        },
        {
            'address': 0xABCD001C,
            'data': b'\x0C\x0D\x0E\x0F'
        }
    ]

    real = ihex.cut_to_pages(input, page_size)
    assert(real == predict)

if __name__ == '__main__':
    test_ihex_parse_1()
    test_ihex_parse_2()
    test_ihex_parse_3()
    test_ihex_padding_1()
    test_ihex_padding_2()
    test_ihex_cut_1()



# -*- coding: utf-8 -*-
"""
Sub-command implementation.
"""

from serprog import device
from serprog import loader
from serprog import exceptions

import serial.tools.list_ports

import progressbar
import serial
import sys

def do_print_devices(args):
    s  = "Available device list:\n"
    s += "    device name   \t num \t note\n"
    for dev in device.device_list:
        s += "  - {0:11s}  \t {1:4s}\t {2}\n".format(
            dev['name'], str(dev['dev_type']), dev['note'])

    print(s)

def do_print_ports(args):
    for (port, desc, hwid) in serial.tools.list_ports.comports():
        print("{:20}".format(port))
        print("    desc: {}".format(desc))
        print("    hwid: {}".format(hwid))

def do_prog(args):

    # Create Serial object
    ser = serial.Serial()
    ser.port = args.port
    ser.baudrate = 115200
    ser.timeout = 1
    try:
        ser.open()
    except:
        print('ERROR: {0} has been opened by another application.'.format(args.port))
        sys.exit(1)

    is_prog_flash     = bool(args.flash_file)
    is_ext_flash      = bool(args.ext_flash_file)
    is_prog_eeprom    = bool(args.eeprom_file)
    is_ext_flash_boot = bool(args.ext_flash_boot)

    # Device number
    device_type = device.get_device_by_str(args.device)

    try:
        l = loader.Loader(
            ser               = ser,
            device_type       = device_type,
            is_flash_prog     = is_prog_flash,
            is_ext_flash_prog = is_ext_flash,
            is_eeprom_prog    = is_prog_eeprom,
            is_ext_flash_boot = is_ext_flash_boot,
            flash_file        = args.flash_file,
            ext_flash_file    = args.ext_flash_file,
            eeprom_file       = args.eeprom_file,
        )
    except exceptions.ComuError:
        print("ERROR: Can't communicate with the device.")
        print("       Please check the comport and the device.")
        sys.exit(1)
    except exceptions.CheckDeviceError as e:
        print("ERROR: Device is not match.")
        print("       Assigned device is '{0:s}'".format(device.device_list[e.in_dev]['name']))
        print("       Detected device is '{0:s}'".format(device.device_list[e.real_dev]['name']))
        sys.exit(1)

    print("Device is '{0:s}'".format(device.device_list[l.device_type]['name']))
    print('Flash hex size is {0:0.2f} KB ({1} bytes)'.format(l.flash_size/1024, l.flash_size))
    print('Externel Flash hex size is {0:0.2f} KB ({1} bytes)'.format(l.ext_flash_size/1024, l.ext_flash_size))
    print('EEPROM hex size is {0} bytes.'.format(l.eeprom_size))
    print('Estimated time  is {0:0.2f} s.'.format(l.prog_time))

    widgets = [
        ' [', progressbar.Timer('Elapsed Time: %(seconds)0.2fs', ), '] ',
        progressbar.Bar(),
        progressbar.Counter(format='%(percentage)0.2f%%'),
    ]

    bar = progressbar.ProgressBar(max_value=l.total_steps, widgets=widgets)
    bar.update(0)
    for i in range(l.total_steps):
        try:
            l.do_step()
            bar.update(i)
        except exceptions.ComuError:
            print("ERROR: Can't communicate with the device.")
            print("Please check the comport is correct.")
            break
        except Exception:
            bar.finish(end='\n', dirty=True)
            print("ERROR: Can't communicate with the device.")
            print("Please check the comport is correct.")
            break

    bar.finish(end='\n')
    ser.close()



# -*- coding: utf-8 -*-
"""
CLI parameter paser
"""

from serprog import device
from serprog import business
from serprog import ihex

import serial.tools.list_ports
import argparse
import os


def parser_init(parser: argparse.ArgumentParser):

    parser.description = 'A serial and secure programming tool for microcontroller.'

    subparsers = parser.add_subparsers(dest='subcmd')

    # parser of 'prog' subcommand
    parser_pr = subparsers.add_parser(
        'prog',
        aliases = [],
        help = 'Program the image to the board.'
    )

    parser_prog_init(parser_pr)

    # parser of 'print-devices' subcommand
    parser_pd = subparsers.add_parser(
        'print-devices',
        aliases = ['pd'],
        help = 'List all available devices.'
    )

    # parser of 'print-ports' subcommand
    parser_pp = subparsers.add_parser(
        'print-ports',
        aliases = ['pp'],
        help = 'List all available serial ports.'
    )


def parser_prog_init(parser: argparse.ArgumentParser):
    """ Parser of CLI sub-command.

    Args:
        parser (argparse.ArgumentParser): Parser of CLI sub-command
    """

    ## Select device type. -d
    arg_d_help = 'The name or number of the device type to be programmed. '
    arg_d_help += 'Can see available device type by subcommand print-device-list.'
    parser.add_argument(
        *('-d', '--decice'),
        action      = 'store',
        dest        = 'device',
        type        = str,
        default     = 'auto',
        help        = arg_d_help
    )

    ## Select serial com port. -p
    arg_p_help = 'The serial port which program burn the device.'
    parser.add_argument(
        *('-p', '--port'),
        action      = 'store',
        dest        = 'port',
        type        = str,
        required    = True,
        help        = arg_p_help
    )

    # Select programmed flash image. -f
    arg_f_help = 'Set binary file which program to flash.'
    parser.add_argument(
        *('-f', '--flash'),
        action      = 'store',
        dest        = 'flash_file',
        type        = str,
        required    = False,
        help        = arg_f_help
    )

    # Select programmed external flash image. -ef
    arg_ef_help = 'Set binary file which program to externel flash.'
    parser.add_argument(
        *('-ef', '--extflash'),
        action      = 'store',
        dest        = 'ext_flash_file',
        type        = str,
        required    = False,
        help        = arg_ef_help
    )

    # Select programmed eeprom image. -e
    arg_e_help = 'Set binary file which program to eeprom.'
    parser.add_argument(
        *('-e', '--eeprom'),
        action      = 'store',
        dest        = 'eeprom_file',
        type        = str,
        required    = False,
        help        = arg_e_help
    )

    # Select programmed file name in embedded file system. -flashboot
    arg_flash_boot_help = 'Program from externel flash to internel flash.'
    parser.add_argument(
        *('-flashboot', '--extflash_boot'),
        action      = 'store_true',
        dest        = 'ext_flash_boot',
        # type        = str,
        required    = False,
        help        = arg_flash_boot_help
    )


def chk_prog_args(args: argparse.Namespace) -> bool:
    """ Check the 'prog' sub-command.

    Args:
        args (argparse.Namespace): CLI paser command.

    Returns:
        bool: True, legal; False, illegal.
    """

    ## Select device type. -d
    res = device.get_device_by_str(args.device)

    if res == -1:
        print('Error: Parameter --device is illegal.')
        business.do_print_devices(args)
        return False

    # Check whether it is ext_flash_boot status.
    if (args.ext_flash_boot):
        pass
    elif (args.flash_file is None) and (args.eeprom_file is None) and (args.ext_flash_file is None):
        errmsg = 'Error: No flash or eeprom needs to be burned, please use \'-f \', \'-e \', \'-E \' to specify the file.'
        print(errmsg)
        return False

    # Flash file check.
    if args.flash_file is not None:
        if not os.path.isfile(args.flash_file):
            errmsg = 'Error: Cannot find flash binary file {0}.'
            print(errmsg.format(args.flash_file))
            return False
        elif not ihex.is_ihex(args.flash_file):
            errmsg = 'Error: The flash binary file {0} is not ihex formatted.'
            print(errmsg.format(args.flash_file))
            return False

    # External flash file check.
    if args.ext_flash_file is not None:
        if not os.path.isfile(args.ext_flash_file):
            errmsg = 'Error: Cannot find flash binary file {0}.'
            print(errmsg.format(args.ext_flash_file))
            return False
        elif not ihex.is_ihex(args.ext_flash_file):
            errmsg = 'Error: The flash binary file {0} is not ihex formatted.'
            print(errmsg.format(args.ext_flash_file))
            return False

    # EEPROM file check.
    if args.eeprom_file is not None:
        if not os.path.isfile(args.eeprom_file):
            errmsg = 'Error: Cannot find eeprom binary file {0}.'
            print(errmsg.format(args.eeprom_file))
            return False
        elif not ihex.is_ihex(args.eeprom_file):
            errmsg = 'Error: The eeprom binary file {0} is not ihex formatted.'
            print(errmsg.format(args.eeprom_file))
            return False

    # Serial port check.
    if args.port not in [p[0] for p in serial.tools.list_ports.comports()]:
        print('Error: Cannot find serial port {0}.'.format(args.port))
        print('The available serial ports are as follows:')
        business.do_print_ports(args)
        return False

def chk_print_ports_args(args: argparse.Namespace) -> bool:
    """ Check the 'print-ports' sub-command.

    Args:
        args (argparse.Namespace): CLI paser command.

    Returns:
        bool: True, legal; False, illegal.
    """
    # No check ...
    return True

def chk_print_devices_args(args: argparse.Namespace) -> bool:
    """ Check the 'print-devices' sub-command.

    Args:
        args (argparse.Namespace): CLI paser command.

    Returns:
        bool: True, legal; False, illegal.
    """
    # No check ...
    return True



# -*- coding: utf-8 -*-

import serprog
import setuptools

with open('README.md', 'r') as f:
    long_description = f.read()

with open('requirements.txt','r') as f:
    install_requires = f.read().split('\n')
    install_requires.remove('')

setuptools.setup(
    name = 'serprog',
    version = serprog.__version__,
    description = 'A serial and secure programming tool for microcontroller.',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    author = 'cy023',
    author_email = 'cyyang023@gmail.com',
    url = 'https://github.com/cy023/SerProg',
    license = 'MIT',
    packages = ['serprog'],
    zip_safe = False,
    entry_points = {
        'console_scripts': [
            'serprog = serprog.__main__:run'
        ],
    },
    install_requires = install_requires
)



# -*- coding: utf-8 -*-
"""
serprog CLI entry.
"""

from serprog import cmdline
from serprog import business

import argparse
import sys

def run():
    # Parse CLI command
    parser = argparse.ArgumentParser()
    cmdline.parser_init(parser)
    args = parser.parse_args()

    # Parse CLI sub-command
    if args.subcmd == 'print-devices' or args.subcmd == 'pd':
        if cmdline.chk_print_devices_args(args) is False:
            sys.exit(1)
        else:
            business.do_print_devices(args)

    elif args.subcmd == 'print-ports' or args.subcmd == 'pp':
        if cmdline.chk_print_ports_args(args) is False:
            sys.exit(1)
        else:
            business.do_print_ports(args)

    elif args.subcmd == 'prog':
        if cmdline.chk_prog_args(args) is False:
            sys.exit(1)
        else:
            business.do_prog(args)

    else:
        parser.print_help()


if __name__ == '__main__':
    run()



# -*- coding: utf-8 -*-
""" Loader API
CommandTrnasHandler Object is internal object of Loader.
"""

import enum
import os
import time
import serial
import sys
import datetime
from pathlib import Path

from serprog import bootprotocol
from serprog import device
from serprog import exceptions
from serprog import ihex

from typing import Union

__all__ = ['Loader']


class CommandTrnasHandler():
    """ Manager of Command Response.

    The main task.
        1. Send command
        2. wait
        3. Receive response

    You can use the functions `cmd_xxx` to send commands and receive responses.
    For the parameters of each command, see the function and bootprotocol specifications.
    """

    def __init__(self, ser: serial.Serial):
        """ Initialization
        Args:
            ser (serial.Serial): The serial object used for communication must be opened by the outside first.
        """
        self._ser = ser
        self._pd = bootprotocol.Decoder()
        self.timeout = 5

    def _get_packet(self):
        """ Get Packet function (polling)

        Raises:
            exceptions.ComuError: Timeout or packet format error.

        Returns:
            [dict[serprog.alp.Command, bytearray]]: Packet object.

        Packet object format:
            res = {
                'command': (serprog.alp.Command) command number,
                'data': (bytearray) packet data.
            }
        """
        if sys.platform == 'win32':
            start_time = time.perf_counter()
        elif sys.platform == 'linux':
            start_time = time.process_time()

        exec_time = 0
        exit_flag = False
        pac_decode_err_flag = False
        packet = None

        while exec_time < 3 and exit_flag is False:
            ch = self._ser.read(1)

            if len(ch) != 0:
                self._pd.step(ch[0])

            if self._pd.isDone():
                packet = self._pd.getPacket()
                exit_flag = True
            elif self._pd.isError():
                pac_decode_err_flag = True
                exit_flag = True

            if sys.platform == 'win32':
                exec_time = time.perf_counter() - start_time
            elif sys.platform == 'linux':
                exec_time = time.process_time() - start_time

        if exec_time > self.timeout:
            # TODO: timeout error, packet format error
            raise exceptions.ComuError

        if pac_decode_err_flag:
            # packet decode error
            raise exceptions.ComuError
        # print('\033[93m' + '[_get_packet]' + '\033[0m', packet)
        return packet

    def _block_get_packet(self):
        """ Get Packet function (blocking mode), for the long time operation.
        """
        packet = None

        while True:
            ch = self._ser.read(1)

            if len(ch) != 0:
                self._pd.step(ch[0])

            if self._pd.isDone():
                packet = self._pd.getPacket()
                break
            elif self._pd.isError():
                raise exceptions.ComuError
        # print('\033[93m' + '[_block_get_packet]' + '\033[0m', packet)
        return packet

    def _put_packet(self, cmd: Union[bootprotocol.CMD, int], data: bytearray):
        """ Put Packet function (polling)

        Args:
            cmd (Union[alp.Command, int]): command number.
            data (bytearray): packet data.
        """
        req_raw = bootprotocol.encode(cmd, data)
        # print('\033[93m' + '\n[_put_packet]' + '\033[0m', req_raw)
        self._ser.write(req_raw)

    ###############################

    def cmd_chk_protocol(self):
        self._put_packet(bootprotocol.CMD.CHK_PROTOCOL, b'test')
        res = self._get_packet()

        if res == None:
            return False, 0
        elif res['command'] == bootprotocol.CMD.CHK_PROTOCOL and res['data'][0] == 0:
            return True, res['data'][1]
        else:
            return False, 0

    def cmd_chk_device(self):
        self._put_packet(bootprotocol.CMD.CHK_DEVICE, b'')
        res = self._get_packet()
        if res['command'] == bootprotocol.CMD.CHK_DEVICE and res['data'][0] == 0:
            return True, res['data'][1]
        else:
            return False, int(0)

    def cmd_prog_end(self):
        self._put_packet(bootprotocol.CMD.PROG_END, b'')
        res = self._get_packet()
        return res['command'] == bootprotocol.CMD.PROG_END and res['data'][0] == 0

    def cmd_prog_ext_flash_boot(self):
        self._put_packet(bootprotocol.CMD.PROG_EXT_FLASH_BOOT, b'')
        res = self._block_get_packet() # waiting for a while ...
        return res['command'] == bootprotocol.CMD.PROG_EXT_FLASH_BOOT and res['data'][0] == 0

    def cmd_flash_set_pgsz(self, size):
        self._put_packet(bootprotocol.CMD.FLASH_SET_PGSZ, size.to_bytes(4, 'little'))
        res = self._get_packet()
        return res['command'] == bootprotocol.CMD.FLASH_SET_PGSZ and res['data'][0] == 0

    def cmd_flash_get_pgsz(self):
        self._put_packet(bootprotocol.CMD.FLASH_GET_PGSZ, b'')
        res = self._get_packet()
        if res['command'] == bootprotocol.CMD.FLASH_GET_PGSZ and res['data'][0] == 0:
            return True, int.from_bytes(res['data'][1:3], 'little')
        else:
            return False, int(0)

    ###############################

    def cmd_flash_write(self, page_addr, data):
        paylod = page_addr.to_bytes(4, 'little') + data
        self._put_packet(bootprotocol.CMD.FLASH_WRITE, paylod)
        res = self._get_packet()
        return res['command'] == bootprotocol.CMD.FLASH_WRITE and res['data'][0] == 0

    def cmd_flash_read(self):
        self._put_packet(bootprotocol.CMD.FLASH_READ, b'')
        res = self._get_packet()
        if res['command'] == bootprotocol.CMD.FLASH_READ and res['data'][0] == 0:
            return True, res['data']
        else:
            return False, bytearray(b'')

    def cmd_flash_earse_sector(self, num):
        self._put_packet(bootprotocol.CMD.FLASH_EARSE_SECTOR,
                         num.to_bytes(2, 'little'))
        res = self._get_packet()
        if res['command'] == bootprotocol.CMD.FLASH_EARSE_SECTOR and res['data'][0] == 0:
            return True, int.from_bytes(res['data'][1:5], 'little')
        else:
            return False, int(0)
        
    def cmd_flash_earse_all(self):
        self._put_packet(bootprotocol.CMD.FLASH_EARSE_ALL, b'')
        res = self._block_get_packet() # waiting for a while ...
        return res['command'] == bootprotocol.CMD.FLASH_EARSE_ALL and res['data'][0] == 0

    ###############################

    def cmd_ext_flash_fopen(self):
        self._put_packet(bootprotocol.CMD.EXT_FLASH_FOPEN, b'fopen')
        res = self._get_packet()
        return res['command'] == bootprotocol.CMD.EXT_FLASH_FOPEN and res['data'][0] == 0

    def cmd_ext_flash_write(self, page_addr, data):
        paylod = page_addr.to_bytes(4, 'little') + data
        self._put_packet(bootprotocol.CMD.EXT_FLASH_WRITE, paylod)
        res = self._get_packet()
        return res['command'] == bootprotocol.CMD.EXT_FLASH_WRITE and res['data'][0] == 0

    def cmd_ext_flash_read(self):
        pass

    def cmd_ext_flash_fclose(self):
        now = datetime.datetime.now()
        t_stamp = bytearray([now.minute, now.hour, now.day, now.month, (now.year - 2000)])
        self._put_packet(bootprotocol.CMD.EXT_FLASH_FCLOSE, t_stamp)
        res = self._get_packet()
        if res['command'] == bootprotocol.CMD.EXT_FLASH_FCLOSE and res['data'][0] == 0:
            return True
        else:
            return False

    ###############################

    def cmd_eeprom_set_pgsz(self, size):
        self._put_packet(bootprotocol.CMD.EEPROM_SET_PGSZ, size.to_bytes(4, 'little'))
        res = self._get_packet()
        return res['command'] == bootprotocol.CMD.EEPROM_SET_PGSZ and res['data'][0] == 0

    def cmd_eeprom_get_pgsz(self):
        self._put_packet(bootprotocol.CMD.EEPROM_GET_PGSZ, b'')
        res = self._get_packet()
        if res['command'] == bootprotocol.CMD.EEPROM_GET_PGSZ and res['data'][0] == 0:
            return True, int.from_bytes(res['data'][1:3], 'little')
        else:
            return False, int(0)

    def cmd_eeprom_write(self, page_data):
        self._put_packet(bootprotocol.CMD.EEPROM_WRITE, page_data)
        res = self._get_packet()
        if res['command'] == bootprotocol.CMD.EEPROM_WRITE and res['data'][0] == 0:
            return True, int.from_bytes(res['data'][1:5], 'little')
        else:
            return False, int(0)

    def cmd_eeprom_read(self):
        self._put_packet(bootprotocol.CMD.EEPROM_READ, b'')
        res = self._get_packet()
        if res['command'] == bootprotocol.CMD.EEPROM_READ and res['data'][0] == 0:
            return True, int.from_bytes(res['data'][1:5], 'little')
        else:
            return False, int(0)

    def cmd_eeprom_earse(self):
        self._put_packet(bootprotocol.CMD.EEPROM_EARSE, b'')
        res = self._get_packet()
        if res['command'] == bootprotocol.CMD.EEPROM_EARSE and res['data'][0] == 0:
            return True, int.from_bytes(res['data'][1:5], 'little')
        else:
            return False, int(0)

    def cmd_eeprom_earse_all(self):
        self._put_packet(bootprotocol.CMD.EEPROM_EARSE_ALL, b'')
        res = self._get_packet()
        return res['command'] == bootprotocol.CMD.EEPROM_EARSE_ALL and res['data'][0] == 0


class Loader():
    """ Programming transaction management object.

    Raises:
        exceptions.CheckDeviceError: Device can't be dectected.
        exceptions.DeviceTypeError: The detected device is a different type than the specified device.
        FileNotFoundError: Can't find flash or eeprom programming file (image).
        exceptions.ComuError: Communication error.
        exceptions.FlashIsNotIhexError: The flash programming file (image) is not intel hex format.
        exceptions.EepromIsNotIhexError: The eeprom programming file (image) is not intel hex format.
    """
    _device_type       = int()
    _device_name       = str()

    _is_flash_prog     = bool()
    _is_ext_flash_prog = bool()
    _is_ext_flash_boot = bool()
    _is_eeprom_prog    = bool()

    _flash_file        = str()
    _ext_flash_file    = str()
    _eeprom_file       = str()

    class _Stage(enum.IntEnum):
        """ Programming transaction management object status
        """
        PREPARE        = 0  # Preparing. check image file, parameter ...
        FLASH_PROG     = 1  # flash programming ...
        EEPROM_PROG    = 2  # eeprom programming ...
        EXT_FLASH_PROG = 3  # external flash programming ...
        EXT_FLASH_BOOT = 4  # external to internel programming ...
        END            = 5  # Finish. Send 'END' or 'JUMP to APP' command.

    _stage = _Stage(_Stage.PREPARE)
    _stage_iter = None

    _total_steps = 0
    _cur_step    = 0

    _flash_pages        = list()
    _flash_page_idx     = int()
    _ext_flash_pages    = list()
    _ext_flash_page_idx = int()
    _eeprom_pages       = list()
    _eeprom_page_idx    = int()

    # output info
    _flash_size     = int(0)
    _ext_flash_size = int(0)
    _eeprom_size    = int(0)
    _prog_time      = float(0)

    def __init__(
        self,
        ser:                serial.Serial,
        device_type:        int = 0,
        is_flash_prog:      bool = False,
        is_ext_flash_prog:  bool = False,
        is_eeprom_prog:     bool = False,
        is_ext_flash_boot:  bool = False,
        flash_file:         str = '',
        ext_flash_file:     str = '',
        eeprom_file:        str = '',
    ):
        """ Initialization

        Args:
            ser (serial.Serial):
                The serial object used for communication must be opened by the outside first.
            device_type (int, optional):
                Device type. The default is 0.
            is_flash_prog (bool, optional):
                Whether to program flash. The default is False.
            is_ext_flash_prog (bool, optional):
                Whether to program external flash. The default is False.
            is_eeprom_prog (bool, optional):
                Whether to program eeprom. The default is False.
            is_ext_flash_boot (bool, optional):
                Whether to burn files from external flash to internal flash. The default is False.
            flash_file (str, optional):
                The flash image. The default is ''.
            eeprom_file (str, optional):
                The eeprom image. The default is ''.
        """
        self._ser = ser

        self._cth = CommandTrnasHandler(ser)

        self._device_type       = device_type
        self._is_flash_prog     = is_flash_prog
        self._is_ext_flash_prog = is_ext_flash_prog
        self._is_eeprom_prog    = is_eeprom_prog
        self._is_ext_flash_boot = is_ext_flash_boot
        self._flash_file        = flash_file
        self._ext_flash_file    = ext_flash_file
        self._eeprom_file       = eeprom_file

        self._prepare()

        self._prog_time

    @property
    def stage(self):
        return self._stage

    @property
    def device_type(self):
        return self._device_type

    @property
    def device_name(self):
        return self._device_name

    @property
    def total_steps(self):
        return self._total_steps

    @property
    def flash_size(self):
        return self._flash_size

    @property
    def ext_flash_size(self):
        return self._ext_flash_size

    @property
    def eeprom_size(self):
        return self._eeprom_size

    @property
    def prog_time(self):
        return self._prog_time

    @property
    def ext_flash_file(self):
        return Path(self._ext_flash_file).stem

    def _prepare(self):
        """ Preparation function before programming.

        1. Check parameters
        2. Check the flash and eeprom burning files
        3. Detection device
        4. Generate action list
        """
        if self._device_type > len(device.device_list):
            raise exceptions.DeviceTypeError(self._device_type)

        if self._is_flash_prog:
            if os.path.isfile(self._flash_file) is False:
                raise FileNotFoundError

        if self._is_ext_flash_prog:
            if os.path.isfile(self._ext_flash_file) is False:
                raise FileNotFoundError

        if self._is_eeprom_prog:
            if os.path.isfile(self._eeprom_file) is False:
                raise FileNotFoundError

        self._prepare_flash()
        self._prepare_ext_flash()
        self._prepare_eeprom()
        self._prepare_device()

        # Stage
        stg_list = list()
        if self._is_flash_prog:
            stg_list.append(self._Stage.FLASH_PROG)
            self._total_steps += len(self._flash_pages)
        if self._is_ext_flash_prog:
            stg_list.append(self._Stage.EXT_FLASH_PROG)
            self._total_steps += len(self._ext_flash_pages)
        if self._is_eeprom_prog:
            stg_list.append(self._Stage.EEPROM_PROG)
            self._total_steps += len(self._eeprom_pages)
        if self._is_ext_flash_boot:
            stg_list.append(self._Stage.EXT_FLASH_BOOT)
        stg_list.append(self._Stage.END)
        self._total_steps += 1
        self._stage_iter = iter(stg_list)
        self._stage = next(self._stage_iter)

        # prog time
        if self._device_type == 1: # D_ATSAME54_DEVB
            self._prog_time = len(self._flash_pages) * \
                0.5 + len(self._eeprom_pages) * 0.05 + 0.23

    def _prepare_device(self):
        """ Check if the device matches the set device number.

        Raises:
            exceptions.ComuError: Unable to communicate
            exceptions.CheckDeviceError: Device comparison error.
        """
        res, self._protocol_version = self._cth.cmd_chk_protocol()

        if res and self._protocol_version == 1:
            res2, detected_device = self._cth.cmd_chk_device()
            if res2 is False:
                raise exceptions.ComuError()
        else:
            raise exceptions.ComuError()

        # auto detect device
        if device.device_list[self._device_type]['protocol_version'] == 0:
            self._device_type = detected_device

        elif device.device_list[self._device_type]['protocol_version'] == 1:
            if detected_device != self._device_type:
                raise exceptions.CheckDeviceError(self._device_type, detected_device)

        self._device_name = device.device_list[self._device_type]['name']

    def _prepare_flash(self):
        """ Process flash programming file

        Raises:
            exceptions.FlashIsNotIhexError: flash image file is not in intel hex format.

        1. Detect whether it is intel hex format
        2. Take out the data
        3. padding_space
        4. cut_to_pages
        """
        if self._is_flash_prog:
            try:
                blocks = ihex.parse(self._flash_file)
                self._flash_size = sum([len(block['data']) for block in blocks])
                blocks = ihex.padding_space(blocks, 512, bytes.fromhex('FF'))
                self._flash_pages = ihex.cut_to_pages(blocks, 512)
            except Exception:
                raise exceptions.FlashIsNotIhexError(self._flash_file)
    
    def _prepare_ext_flash(self):
        """ Process external flash programming files

        The basic operation is the same as _prepare_flash, and the difference will only be made when sending subsequent packets.
        """
        if self._is_ext_flash_prog:
            try:
                blocks = ihex.parse(self._ext_flash_file)
                self._ext_flash_size = sum([len(block['data']) for block in blocks])
                blocks = ihex.padding_space(blocks, 512, bytes.fromhex('FF'))
                self._ext_flash_pages = ihex.cut_to_pages(blocks, 512)
            except Exception:
                raise exceptions.FlashIsNotIhexError(self._ext_flash_file)

    def _prepare_eeprom(self):
        # TODO Discuss eeprom programming specifications, and complete.
        """ Process eeprom programming file.

        Raises:
            exceptions.EepromIsNotIhexError: eeprom image file is not in intel hex format

        1. Detect whether it is intel hex format
        2. Take out the data
        3. padding_space
        4. cut_to_pages
        """
        if self._is_eeprom_prog:
            try:
                blocks = ihex.parse(self._eeprom_file)
                self._eeprom_size = sum([len(block['data']) for block in blocks])
                blocks = ihex.padding_space(blocks, 512, bytes.fromhex('FF'))
                self._eeprom_pages = ihex.cut_to_pages(blocks, 512)
            except Exception:
                raise exceptions.EepromIsNotIhexError(self._eeprom_file)

    def _do_flash_prog_step(self):
        address = self._flash_pages[self._flash_page_idx]['address']
        data = self._flash_pages[self._flash_page_idx]['data']

        if self._flash_page_idx == 0:
            self._cth.cmd_flash_earse_all()
        self._cth.cmd_flash_write(address, data)

        self._flash_page_idx += 1
        self._cur_step += 1

        if self._flash_page_idx == len(self._flash_pages):
            self._stage = next(self._stage_iter)

    def _do_ext_flash_prog_step(self):
        # Programming to external flash, the actual action content is the same as flash_prog
        address = self._ext_flash_pages[self._ext_flash_page_idx]['address']
        data = self._ext_flash_pages[self._ext_flash_page_idx]['data']

        if self._ext_flash_page_idx == 0:
            self._cth.cmd_ext_flash_fopen()
        self._cth.cmd_ext_flash_write(address, data)

        self._ext_flash_page_idx += 1
        self._cur_step += 1

        if self._ext_flash_page_idx == len(self._ext_flash_pages):
            self._cth.cmd_ext_flash_fclose()
            self._stage = next(self._stage_iter)

    def _do_ext_flash_boot_step(self):
        # Send external flash programming to internal flash command
        self._cth.cmd_prog_ext_flash_boot()

    def _do_eeprom_prog_step(self):
        self._cth.cmd_eeprom_write(self._eeprom_pages[self._eeprom_page_idx])

        self._eeprom_page_idx += 1
        self._cur_step += 1
        if self._eeprom_page_idx == len(self._eeprom_pages):
            self._stage = next(self._stage_iter)

    def _do_prog_end_step(self):
        self._cth.cmd_prog_end()
        self._cur_step += 1

    def do_step(self):
        if self.stage == self._Stage.FLASH_PROG:
            self._do_flash_prog_step()
        elif self.stage == self._Stage.EEPROM_PROG:
            self._do_eeprom_prog_step()
        elif self.stage == self._Stage.EXT_FLASH_PROG:
            self._do_ext_flash_prog_step()
        elif self.stage == self._Stage.EXT_FLASH_BOOT:
            self._do_ext_flash_boot_step()
        elif self.stage == self._Stage.END:
            self._do_prog_end_step()
